#! /usr/env/python
"""

This tests the overland flow and shear stress generator.

"""

import landlab
from landlab.components.overland_flow.generate_overland_flow import OverlandFlow
from landlab.components.uniform_precip.generate_uniform_precip import PrecipitationDistribution
#import pylab as pl
from pylab import plot, draw, show, contour, imshow, colorbar
import numpy as np

def main():
    """
    do some stuff
    """
    
    # User-defined parameter values
    nr = 5
    nc = 6
    nnodes = nr*nc
    dx=1
    #instantiate grid
    rg = landlab.RasterModelGrid(nr, nc, dx)
    #rg.set_inactive_boundaries(False, False, True, True)
    
    nodata_val=0
    elevations  = nodata_val*np.ones( nnodes )    
    #set-up interior elevations with random numbers
    #for i in range(0, nnodes):
    #    if rg.is_interior(i):
    #        elevations[i]=random.random_sample()
    
    #set-up with prescribed elevations to test drainage area calcualtion
    helper = [7,8,9,10,13,14,15,16]
    elevations[helper]=2
    helper = [19,20,21,22]
    elevations[helper]=3        
    elevations[7]=1    
    
    # Get a 2D array version of the elevations
    elev_raster = rg.node_vector_to_raster(elevations,True)
    
    of=OverlandFlow('input_data.txt',rg,0)
    rainfall = PrecipitationDistribution()
    rainfall.initialize('input_data.txt')
    rainfall.update
    
    #for now this is in hours, so put into seconds
    storm_duration = rainfall.storm_duration*3600
    #in mm/hour, so convert to m/second
    storm_intensity = rainfall.intensity/1000/3600
    print "storm duration, seconds ", storm_duration
    print "storm duration, hours ", rainfall.storm_duration
    print "storm intensity ", storm_intensity
    
    tau = of.run_one_step(rg,elevations,storm_duration,storm_intensity)
    
    #imshow(elev_raster)
    #colorbar()
    #show()
    #
    ## Create a shaded image
    #pylab.close()  # clear any pre-existing plot
    #image_extent = [0, 0.001*dx*nc, 0, 0.001*dx*nr] # in km
    #im = pylab.imshow(elev_raster, cmap=pylab.cm.RdBu, extent=image_extent)
    #pylab.xlabel('Distance (km)', fontsize=12)
    #pylab.ylabel('Distance (km)', fontsize=12)
    
    
if __name__ == "__main__":
    main()


#! /usr/bin/env python
"""
Unit tests for landlab.components.flexure.flexure
"""

import unittest
import numpy as np

from landlab import RasterModelGrid
from landlab.components.flexure.flexure import FlexureComponent


(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e3, 10e3), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


class TestFlexureInfo(unittest.TestCase):
    def setUp(self):
        grid = RasterModelGrid(20, 20, 10e3)
        self.flex = FlexureComponent(grid)

    def test_name(self):
        self.assertEqual(self.flex.name, 'Flexure')

    def test_input_var_names(self):
        self.assertEqual(sorted(self.flex.input_var_names),
                         ['lithosphere__elevation',
                          'lithosphere__overlying_pressure',
                          'planet_surface_sediment__deposition_increment'])

    def test_output_var_names(self):
        self.assertEqual(sorted(self.flex.output_var_names),
                         ['lithosphere__elevation',
                          'lithosphere__elevation_increment'])

    def test_var_units(self):
        self.assertEqual(set(self.flex.input_var_names) |
                         set(self.flex.output_var_names),
                         set(self.flex.units))

        self.assertEqual(
            self.flex.units['lithosphere__elevation'],
            'm')
        self.assertEqual(
            self.flex.units['lithosphere__elevation_increment'],
            'm')
        self.assertEqual(
            self.flex.units['lithosphere__overlying_pressure'],
            'Pa')
        self.assertEqual(
            self.flex.units['planet_surface_sediment__deposition_increment'],
            'm')


class TestFlexureGrid(unittest.TestCase):
    def setUp(self):
        grid = RasterModelGrid(20, 20, 10e3)
        self.flex = FlexureComponent(grid)

    def test_grid_shape(self):
        self.assertEqual(self.flex.grid.number_of_node_rows, _SHAPE[0])
        self.assertEqual(self.flex.grid.number_of_node_columns, _SHAPE[1])

    def test_grid_xdimension(self):
        self.assertEqual(self.flex.grid.get_grid_xdimension(),
                         _SHAPE[1] * _SPACING[1])

    def test_grid_ydimension(self):
        self.assertEqual(self.flex.grid.get_grid_xdimension(),
                         _SHAPE[0] * _SPACING[0])


class TestFlexureFields(unittest.TestCase):
    def setUp(self):
        grid = RasterModelGrid(20, 20, 10e3)
        self.flex = FlexureComponent(grid)

    def test_field_getters(self):
        for name in self.flex.grid['node']:
            field = self.flex.grid['node'][name]
            self.assertIsInstance(field, np.ndarray)
            self.assertEqual(field.shape,
                             (self.flex.grid.number_of_node_rows *
                              self.flex.grid.number_of_node_columns, ))

        with self.assertRaises(KeyError):
            self.flex.grid['not_a_var_name']

    def test_field_initialized_to_zero(self):
        for name in self.flex.grid['node']:
            field = self.flex.grid['node'][name]
            self.assertTrue(np.all(field == 0.))


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python
"""
Unit tests for landlab.io.netcdf module.
"""

import unittest
import os
import numpy as np
from StringIO import StringIO

from landlab.io.netcdf import read_netcdf


_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


class TestReadNetcdf(unittest.TestCase):
    def test_read_netcdf3_64bit(self):
        grid = read_netcdf(os.path.join(_TEST_DATA_DIR, 'test-netcdf3-64bit.nc'))
        self.assertEqual(grid.shape, (4, 3))

    def test_read_netcdf4(self):
        grid = read_netcdf(os.path.join(_TEST_DATA_DIR, 'test-netcdf4.nc'))
        self.assertEqual(grid.shape, (4, 3))


if __name__ == '__main__':
    unittest.main()



#! /usr/env/python
"""

2D numerical model of diffusion, implemented using ModelGrid.
Provides example of an unstructured grid.

Last updated GT August 2013

"""

from landlab import HexModelGrid
import pylab
import time

def main():
    """
    In this simple tutorial example, the main function does all the work: 
    it sets the parameter values, creates and initializes a grid, sets up 
    the state variables, runs the main loop, and cleans up.
    """

    # INITIALIZE
    
    # User-defined parameter values
    numrows = 7          # number of rows in the grid
    basenumcols = 6          # number of columns in the grid
    dx = 10.0             # grid cell spacing
    kd = 0.01             # diffusivity coefficient, in m2/yr
    uplift_rate = 0.001   # baselevel/uplift rate, in m/yr
    num_time_steps = 1000 # number of time steps in run
    output_name = 'hex_grid_example'  # base name for output files
    output_interval = 100 # interval between outputs, in time steps
    
    # Derived parameters
    dt = 0.1*dx**2 / kd    # time-step size set by CFL condition
    next_output = output_interval  # time of next output to file
    
    # Create and initialize a raster model grid
    mg = HexModelGrid(numrows, basenumcols, dx)
    
    # Set up scalar values
    z = mg.create_node_array_zeros()            # node elevations
    dzdt = mg.create_node_array_zeros()  # node rate of elevation change
    
    # Get a list of the interior cells
    interior_cells = mg.get_active_cell_node_ids()

    # Display a message
    print( 'Running diffusion_with_model_grid.py' )
    print( 'Time-step size has been set to ' + str( dt ) + ' years.' )
    start_time = time.time()

    
    # RUN
    
    # Main loop
    for i in range(0, num_time_steps):
        
        # Calculate the gradients and sediment fluxes
        g = mg.calculate_gradients_at_active_links(z)
        qs = -kd*g
        
        # Calculate the net deposition/erosion rate at each node
        dqsds = mg.calculate_flux_divergence_at_nodes(qs)
        
        # Calculate the total rate of elevation change
        dzdt = uplift_rate - dqsds
            
        # Update the elevations
        z[interior_cells] = z[interior_cells] + dzdt[interior_cells] * dt
        
   
    # FINALIZE
    
    import numpy

    # Test solution for a 1-cell hex grid
    if mg.number_of_nodes==7:  # single cell with 6 boundaries
        perimeter = dx*6*(numpy.sqrt(3.)/3.)
        flux = kd*(numpy.amax(z)/dx)
        total_outflux = perimeter*flux
        total_influx = mg.active_cell_areas*uplift_rate
        print 'total influx=',total_influx,'total outflux=',total_outflux
    print('Run time = '+str(time.time()-start_time))
    
    # Plot the points, colored by elevation
    pylab.figure()
    maxelev = numpy.amax(z)
    for i in range(mg.number_of_nodes):
        mycolor = str(z[i]/maxelev)
        pylab.plot(mg.node_x[i], mg.node_y[i], 'o', color=mycolor, ms=50)
    pylab.show()
    
    mg.display_grid()
    

if __name__ == "__main__":
    main()


import numpy as np

from .base import INACTIVE_BOUNDARY


_VALID_ROUTING_METHODS = set(['d8', 'd4'])


def assert_valid_routing_method(method):
    if method not in _VALID_ROUTING_METHODS:
        raise ValueError(
            '%s: routing method not understood. should be one of %s' %
            (method, ', '.join(_VALID_ROUTING_METHODS)))


def _make_optional_arg_into_array(number_of_elements, *args):
    assert(len(args) < 2)
    if len(args) == 0:
        ids = np.arange(number_of_elements)
    else:
        ids = args[0]
        if not isinstance(ids, list) or not isinstance(ids, np.ndarray):
            try:
                ids = list(ids)
            except TypeError:
                ids = [ids]
    return ids


def calculate_gradient_across_cell_faces(grid, node_values, *args, **kwds):
    """calculate_gradient_across_cell_faces(grid, node_values, [cell_ids], out=None)
    """
    cell_ids = _make_optional_arg_into_array(grid.number_of_cells, *args)
    node_ids = grid.node_index_at_cells[cell_ids]

    values_at_neighbors = node_values[grid.get_neighbor_list(node_ids)]
    values_at_nodes = node_values[node_ids].reshape(len(node_ids), 1)

    out = np.subtract(values_at_nodes, values_at_neighbors, **kwds)
    out *= 1. / grid.node_spacing

    return out


def calculate_gradient_across_cell_corners(grid, node_values, *args, **kwds):
    """calculate_gradient_across_cell_corners(grid, node_values, [cell_ids], out=None)
    """
    cell_ids = _make_optional_arg_into_array(grid.number_of_cells, *args)
    node_ids = grid.node_index_at_cells[cell_ids]

    values_at_diagonals = node_values[grid.get_diagonal_list(node_ids)]
    values_at_nodes = node_values[node_ids].reshape(len(node_ids), 1)

    out = np.subtract(values_at_nodes, values_at_diagonals, **kwds)
    np.divide(out, np.sqrt(2.) * grid.node_spacing, out=out)

    return out


def calculate_max_gradient_across_adjacent_cells(grid, node_values, *args,
                                                 **kwds):
    """calculate_max_gradient_across_adjacent_cells(grid, node_values, [cell_ids], method='d4', out=None)

    Calculate the slopes of *node_values*, given at every node in the grid,
    relative to the nodes centered at *cell_ids*. Note that downward slopes
    are reported as positive. That is, the gradient is positive if a neighbor
    node's value is less than that of the node as *cell_ids*.

    If *cell_ids* is not provided, calculate the maximum gradient for all
    cells in the grid.

    The default is to only consider neighbor cells to the north, south, east,
    and west. To also consider gradients to diagonal nodes, set the *method*
    keyword to *d8* (the default is *d4*).

    Use the *out* keyword if you have an array that you want to put the result
    into. If not given, create a new array.

    Use the *return_node* keyword to also the node id of the node in the
    direction of the maximum gradient.

    >>> import landlab
    >>> rmg = landlab.RasterModelGrid(3, 3)
    >>> node_values = rmg.zeros()
    >>> node_values[1] = -1
    >>> calculate_max_gradient_across_adjacent_cells(rmg, node_values, 0)
    array([ 1.])

    Get both the maximum gradient and the node to which the gradient is
    measured.

    >>> calculate_max_gradient_across_adjacent_cells(rmg, node_values, 0, return_node=True)
    (array([ 1.]), array([1]))
    """
    method = kwds.pop('method', 'd4')
    assert_valid_routing_method(method)

    if method == 'd4':
        return calculate_max_gradient_across_cell_faces(
            grid, node_values, *args, **kwds)
    elif method == 'd8':
        neighbor_grads = calculate_max_gradient_across_cell_faces(
            grid, node_values, *args, **kwds)
        diagonal_grads = calculate_max_gradient_across_cell_corners(
            grid, node_values, *args, **kwds)

        return_node = kwds.pop('return_node', False)

        if not return_node:
            return np.choose(neighbor_grads >= diagonal_grads,
                             (diagonal_grads, neighbor_grads), **kwds)
        else:
            max_grads = np.choose(neighbor_grads[0] >= diagonal_grads[0],
                                  (diagonal_grads[0], neighbor_grads[0]),
                                  **kwds)
            node_ids = np.choose(neighbor_grads[0] >= diagonal_grads[0],
                             (diagonal_grads[1], neighbor_grads[1]),
                             **kwds)
            return (max_grads, node_ids)


def calculate_max_gradient_across_cell_corners(grid, node_values, *args,
                                               **kwds):
    """calculate_max_gradient_across_cell_corners(grid, node_values [, cell_ids], return_node=False, out=None)
    """
    return_node = kwds.pop('return_node', False)

    cell_ids = _make_optional_arg_into_array(grid.number_of_cells, *args)

    grads = calculate_gradient_across_cell_corners(grid, node_values, cell_ids)

    if return_node:
        ind = np.argmax(grads, axis=1)
        node_ids = grid.diagonal_cells[grid.node_index_at_cells[cell_ids], ind]
        if 'out' not in kwds:
            out = np.empty(len(cell_ids), dtype=grads.dtype)
        out[:] = grads[xrange(len(cell_ids)), ind]
        return (out, node_ids)
        #return (out, 3 - ind)
    else:
        return grads.max(axis=1, **kwds)


def calculate_max_gradient_across_cell_faces(grid, node_values, *args, **kwds):
    """calculate_max_gradient_across_cell_faces(grid, node_values, [cell_ids], return_node=False, out=None)

    This method calculates the gradients in *node_values* across all four
    faces of the cell or cells with ID *cell_ids*. Slopes downward from the
    cell are reported as positive. If *cell_ids* is not given, calculate
    gradients for all cells.

    Use the *return_node* keyword to return a tuple, with the first element
    being the gradients and the second the node id of the node in the direction
    of the maximum gradient.

    >>> from landlab import RasterModelGrid
    >>> rmg = RasterModelGrid(3, 3)
    >>> values_at_nodes = np.arange(9.)
    >>> calculate_max_gradient_across_cell_faces(rmg, values_at_nodes)
    array([ 3.])
    >>> (_, ind) = calculate_max_gradient_across_cell_faces(rmg, values_at_nodes, return_node=True)
    >>> ind
    array([1])
    """
    return_node = kwds.pop('return_node', False)

    cell_ids = _make_optional_arg_into_array(grid.number_of_cells, *args)

    grads = calculate_gradient_across_cell_faces(grid, node_values, cell_ids)

    if return_node:
        ind = np.argmax(grads, axis=1)
        node_ids = grid.neighbor_nodes[grid.node_index_at_cells[cell_ids], ind]
        if 'out' not in kwds:
            out = np.empty(len(cell_ids), dtype=grads.dtype)
        out[:] = grads[xrange(len(cell_ids)), ind]
        return (out, node_ids)
        #return (out, 3 - ind)
    else:
        return grads.max(axis=1, **kwds)


def active_link_id_of_cell_neighbor(grid, inds, *args):
    """ active_link_id_of_cell_neighbor(grid, link_ids [, cell_ids])

    Return an array of the active link ids for neighbors of *cell_id* cells.
    *link_ids* is an index into the links of a cell as measured
    clockwise starting from the south.

    If *cell_ids* is not given, return neighbors for all cells in the grid.
    """
    cell_ids = _make_optional_arg_into_array(grid.number_of_cells, *args)
    node_ids = grid.node_index_at_cells[cell_ids]
    links = grid.active_node_links(node_ids).T

    if not isinstance(inds, np.ndarray):
        inds = np.array(inds)

    return links[xrange(len(cell_ids)), inds]


def node_id_of_cell_neighbor(grid, inds, *args):
    """ node_id_of_cell_neighbor(grid, neighbor_ids [, cell_ids])

    Return an array of the node ids for neighbors of *cell_id* cells.
    *neighbor_ids* is an index into the neighbors of a cell as measured
    clockwise starting from the south.

    If *cell_ids* is not given, return neighbors for all cells in the grid.
    """
    cell_ids = _make_optional_arg_into_array(grid.number_of_cells, *args)
    node_ids = grid.node_index_at_cells[cell_ids]
    neighbors = grid.get_neighbor_list(node_ids)

    if not isinstance(inds, np.ndarray):
        inds = np.array(inds)

    return neighbors[xrange(len(cell_ids)), 3 - inds]


def node_id_of_cell_corner(grid, inds, *args):
    """ node_id_of_cell_corner(grid, corner_ids [, cell_ids])

    Return an array of the node ids for diagonal neighbors of *cell_id* cells.
    *corner_ids* is an index into the corners of a cell as measured
    clockwise starting from the southeast.

    If *cell_ids* is not given, return neighbors for all cells in the grid.
    """
    cell_ids = _make_optional_arg_into_array(grid.number_of_cells, *args)
    node_ids = grid.node_index_at_cells[cell_ids]
    diagonals = grid.get_diagonal_list(node_ids)

    if not isinstance(inds, np.ndarray):
        inds = np.array(inds)

    return diagonals[xrange(len(cell_ids)), 3 - inds]


def calculate_flux_divergence_at_nodes(grid, active_link_flux, out=None):
    """
    Same as calculate_flux_divergence_at_active_cells, but works with and
    returns a list of net unit fluxes that corresponds to all nodes, rather
    than just active cells.
    
    Note that we DO compute net unit fluxes at boundary nodes (even though
    these don't have active cells associated with them, and often don't have 
    cells of any kind, because they are on the perimeter). It's up to the 
    user to decide what to do with these boundary values.
    
    Example:

    >>> from landlab import RasterModelGrid
    >>> rmg = RasterModelGrid(4, 5, 1.0)
    >>> u = [0., 1., 2., 3., 0.,
    ...      1., 2., 3., 2., 3.,
    ...      0., 1., 2., 1., 2.,
    ...      0., 0., 2., 2., 0.]
    >>> u = np.array(u)
    >>> grad = rmg.calculate_gradients_at_active_links(u)
    >>> grad
    array([ 1.,  1., -1., -1., -1., -1., -1.,  0.,  1.,  1.,  1., -1.,  1.,
            1.,  1., -1.,  1.])
    >>> flux = -grad    # downhill flux proportional to gradient
    >>> df = rmg.calculate_flux_divergence_at_nodes(flux)
    >>> df
    array([ 0., -1., -1.,  1.,  0., -1.,  2.,  4., -2.,  1., -1.,  0.,  1.,
           -4.,  1.,  0., -1.,  0.,  1.,  0.])
        
    If calculate_gradients_at_nodes is called inside a loop, you can
    improve speed by creating an array outside the loop. For example, do
    this once, before the loop:
        
    >>> df = rmg.zeros(centering='node') # outside loop
    >>> rmg.number_of_nodes
    20
        
    Then do this inside the loop:
        
    >>> df = rmg.calculate_flux_divergence_at_nodes(flux, df)
        
    In this case, the function will not have to create the df array.
    """
    assert (len(active_link_flux) == grid.number_of_active_links), \
           "incorrect length of active_link_flux array"
        
    # If needed, create net_unit_flux array
    if out is None:
        out = grid.empty(centering='node')
    out.fill(0.)
    net_unit_flux = out
        
    assert(len(net_unit_flux) == grid.number_of_nodes)
    
    flux = np.zeros(len(active_link_flux) + 1)
    flux[:len(active_link_flux)] = active_link_flux * grid._dx
    
    net_unit_flux[:] = (
        (flux[grid.node_active_outlink_matrix[0][:]] +
         flux[grid.node_active_outlink_matrix[1][:]]) -
        (flux[grid.node_active_inlink_matrix[0][:]] +
         flux[grid.node_active_inlink_matrix[1][:]])) / grid.cellarea

    return net_unit_flux


# TODO: Functions below here still need to be refactored for speed and to
# conform to the interface standards.

def calculate_max_gradient_across_node(grid, u, cell_id):
    """
    This method calculates the gradients in u across all 4 faces of the 
    cell with ID cell_id, and across the four diagonals. It then returns 
    the steepest (most negative) of these values, followed by its dip 
    direction (e.g.: 0.12, 225). i.e., this is a D8 algorithm. Slopes 
    downward from the cell are reported as positive.
        
    This code is actually calculating slopes, not gradients.  
    The max gradient is the most negative, but the max slope is the most
    positive.  So, this was updated to return the max value, not the 
    min.
        
    GT: Might be possible to speed this up using inlink_matrix and 
    outlink_matrix.
    """
    #We have poor functionality if these are edge cells! Needs an exception
    neighbor_cells = grid.get_neighbor_list(cell_id)
    neighbor_cells.sort()        
    diagonal_cells = []
    if neighbor_cells[0]!=-1:
        diagonal_cells.extend([neighbor_cells[0]-1, neighbor_cells[0]+1])
    if neighbor_cells[3]!=-1:
        diagonal_cells.extend([neighbor_cells[3]-1, neighbor_cells[3]+1])
    slopes = []
    diagonal_dx = np.sqrt(2.) * grid._dx  # Corrected (multiplied grid._dx) SN 05Nov13
    for a in neighbor_cells:
        #ng I think this is actually slope as defined by a geomorphologist,
        #that is -dz/dx and not the gradient (dz/dx)
        #print '\n', cell_id
        #print '\n', a
        single_slope = (u[cell_id] - u[a])/grid._dx
        #print 'cell id: ', cell_id
        #print 'neighbor id: ', a
        #print 'cell, neighbor are internal: ', grid.is_interior(cell_id), grid.is_interior(a)
        #print 'cell elev: ', u[cell_id]
        #print 'neighbor elev: ', u[a]
        #print single_slope
        if not np.isnan(single_slope): #This should no longer be necessary, but retained in case
            slopes.append(single_slope)
        else:
            print 'NaNs present in the grid!'
            
    for a in diagonal_cells:
        single_slope = (u[cell_id] - u[a])/(diagonal_dx)
        #print single_slope
        if not np.isnan(single_slope):
            slopes.append(single_slope)
        else:
            print 'NaNs present in the grid!'
    #print 'Slopes list: ', slopes
    #ng thinks that the maximum slope should be found here, not the 
    #minimum slope, old code commented out.  New code below it.
    #if slopes:
    #    min_slope, index_min = min((min_slope, index_min) for (index_min, min_slope) in enumerate(slopes))
    #else:
    #    print u
    #    print 'Returning NaN angle and direction...'
    #    min_slope = np.nan
    #    index_min = 8
    if slopes:
        max_slope, index_max = max((max_slope, index_max) for (index_max, max_slope) in enumerate(slopes))
    else:
        print u
        print 'Returning NaN angle and direction...'
        max_slope = np.nan
        index_max = 8
    
    # North = Zero Radians  = Clockwise Positive
    angles = [180., 270., 90., 0., 225., 135., 315., 45., np.nan] #This is inefficient 
    
    #ng commented out old code
    #return min_slope, angles[index_min]
    return max_slope, angles[index_max]


def calculate_max_gradient_across_node_d4(self, u, cell_id):
    """
    .. deprecated:: 0.1
        Use :func:`calculate_max_gradient_across_cell_faces` instead

    This method calculates the gradients in u across all 4 faces of the 
    cell with ID cell_id. It then returns 
    the steepest (most negative) of these values, followed by its dip 
    direction (e.g.: 90 180). i.e., this is a D4 algorithm. Slopes 
    downward from the cell are reported as positive.
    
    Note that this is exactly the same as calculate_max_gradient_across_node
    except that this is d4, and the other is d8.
    
    This code is actually calculating slopes, not gradients.  
    The max gradient is the most negative, but the max slope is the most
    positive.  So, this was updated to return the max value, not the 
    min.
    """
    node_id = self.node_index_at_cells[cell_id]
    neighbor_nodes = self.get_neighbor_list(node_id)

    grads = (u[node_id] - u[neighbor_nodes]) / self.node_spacing
    ind = np.argmax(grads)
    _ANGLES = (90., 0., 270., 180.)
    return grads[ind], _ANGLES[ind]

    #We have poor functionality if these are edge cells! Needs an exception
    neighbor_cells = self.get_neighbor_list(cell_id)
    neighbor_cells.sort()
    #print 'Node is internal: ', self.is_interior(cell_id)
    #print 'Neighbor cells: ', neighbor_cells

    slopes = []
    for a in neighbor_cells:
        #ng I think this is actually slope as defined by a geomorphologist,
        #that is -dz/dx and not the gradient (dz/dx)
        if self.node_status[a] != INACTIVE_BOUNDARY:
            single_slope = (u[cell_id] - u[a])/self._dx
        else:
            single_slope = -9999
        #single_slope = (u[cell_id] - u[a])/self._dx
        #print 'cell id: ', cell_id
        #print 'neighbor id: ', a
        #print 'cell, neighbor are internal: ', self.is_interior(cell_id), self.is_interior(a)
        #print 'cell elev: ', u[cell_id]
        #print 'neighbor elev: ', u[a]
        #print single_slope
        #if not np.isnan(single_slope): #This should no longer be necessary, but retained in case
        #    slopes.append(single_slope)
        #else:
        #    print 'NaNs present in the grid!'
        slopes.append(single_slope)
            
    #print 'Slopes list: ', slopes
    #ng thinks that the maximum slope should be found here, not the 
    #minimum slope, old code commented out.  New code below it.
    #if slopes:
    #    min_slope, index_min = min((min_slope, index_min) for (index_min, min_slope) in enumerate(slopes))
    #else:
    #    print u
    #    print 'Returning NaN angle and direction...'
    #    min_slope = np.nan
    #    index_min = 8
    if slopes:
        max_slope, index_max = max((max_slope, index_max) for (index_max, max_slope) in enumerate(slopes))
    else:
        print u
        print 'Returning NaN angle and direction...'
        max_slope = np.nan
        index_max = 4
        
    angles = [180., 270., 90., 0., np.nan] #This is inefficient
    
    #ng commented out old code
    #return min_slope, angles[index_min]
    return max_slope, angles[index_max] 


def calculate_slope_aspect_BFP(xs, ys, zs):
    """
    .. codeauthor:: Katy Barnhart <katherine.barnhart@colorado.edu>

    Fits a plane to the given N points with given *xs*, *ys*, and *zs* values
    using single value decomposition. 
   
    Returns a tuple of (*slope*, *aspect*) based on the normal vector to the 
    best fit plane. 
   
    .. note::
        This function does not check if the points fall on a line, rather
        than a plane.
    """
    if not (len(xs) == len(ys) == len(ys)):
        raise ValueError('array must be the same length')
   
    # step 1: subtract the centroid from the points
    x0 = np.mean(xs)
    y0 = np.mean(ys)
    z0 = np.mean(zs)
   
    x = xs - x0
    y = ys - y0
    z = zs - z0
   
    # step 2: create a 3XN matrix of the points for SVD
    # in python, the unit normal to the best fit plane is
    # given by the third column of the U matrix.
    mat = np.vstack((x, y, z))
    U, s, V = np.linalg.svd(mat)
    normal = U[:, 2]
      
    # step 3: calculate the aspect   
    asp = 90.0 - np.degrees(np.arctan2(normal[1], normal[0]))
    asp = asp % 360.0

    # step 4: calculate the slope   
    slp = 90.0 - np.degrees(np.arcsin(normal[2]))

    return slp, asp



#! /usr/env/python
"""

2D numerical model of diffusion, implemented using ModelGrid.
Provides a simple tutorial example of ModelGrid functionality.

Last updated GT September 2013

"""

from landlab import RasterModelGrid
import pylab, time

def main():
    """
    In this simple tutorial example, the main function does all the work: 
    it sets the parameter values, creates and initializes a grid, sets up 
    the state variables, runs the main loop, and cleans up.
    """
    
    # INITIALIZE
    
    # User-defined parameter values
    numrows = 20          # number of rows in the grid
    numcols = 30          # number of columns in the grid
    dx = 10.0             # grid cell spacing
    kd = 0.01             # diffusivity coefficient, in m2/yr
    uplift_rate = 0.0001  # baselevel/uplift rate, in m/yr
    num_time_steps = 10000 # number of time steps in run
    
    # Derived parameters
    dt = 0.1*dx**2 / kd    # time-step size set by CFL condition
    
    # Create and initialize a raster model grid
    mg = RasterModelGrid(numrows, numcols, dx)
    
    # Set the boundary conditions
    mg.set_inactive_boundaries(False, False, True, True)

    # Set up scalar values
    z = mg.add_zeros('node', 'Elevation')            # node elevations
    
    # Get a list of the interior cells
    interior_cells = mg.get_active_cell_node_ids()

    # Display a message, and record the current clock time
    print( 'Running diffusion_with_model_grid.py' )
    print( 'Time-step size has been set to ' + str( dt ) + ' years.' )
    start_time = time.time()

    # RUN
    
    # Main loop
    for i in range(0, num_time_steps):
        
        # Calculate the gradients and sediment fluxes
        g = mg.calculate_gradients_at_active_links(z)
        qs = -kd*g
        
        # Calculate the net deposition/erosion rate at each node
        dqsds = mg.calculate_flux_divergence_at_nodes(qs)
        
        # Calculate the total rate of elevation change
        dzdt = uplift_rate - dqsds
            
        # Update the elevations
        z[interior_cells] = z[interior_cells] + dzdt[interior_cells] * dt


    # FINALIZE

    # Get a 2D array version of the elevations
    zr = mg.node_vector_to_raster(z, flip_vertically=True)
    
    # Create a shaded image
    pylab.close()  # clear any pre-existing plot
    im = pylab.imshow(zr, cmap=pylab.cm.RdBu, extent=[0,numcols*dx,0,numrows*dx])
    # add contour lines with labels
    cset = pylab.contour(zr, extent=[0,numcols*dx,numrows*dx,0], hold='on')
    pylab.clabel(cset, inline=True, fmt='%1.1f', fontsize=10)
    
    # add a color bar on the side
    cb = pylab.colorbar(im)
    cb.set_label('Elevation in meters')
    
    # add a title and axis labels
    pylab.title('Simulated topography with uplift and diffusion')
    pylab.xlabel('Distance (m)')
    pylab.ylabel('Distance (m)')

    # Display the plot
    pylab.show()
    print('Run time = '+str(time.time()-start_time))

if __name__ == "__main__":
    main()



from landlab.components.uniform_precip.generate_uniform_precip import PrecipitationDistribution


def main():
    print 'We are going to use TrialRun as our class instance.'
    TrialRun = PrecipitationDistribution()
    print 'TrialRun = PrecipitationDistribution()'
    print '\n'

    print "TrialRun's values before initiation..."
    print "Storm Duration is: ", TrialRun.storm_duration, 'hours.'
    print "Interstorm Duration is: ", TrialRun.interstorm_duration, 'hours.'
    print "Storm Depth is: ", TrialRun.storm_depth, 'mm.'
    print "Intensity is: ", TrialRun.intensity, 'mm/hr.' 
    print '\n'
    print 'We should initialize TrialRun... TrialRun.initialize()'
    TrialRun.initialize()
    print '\n'
    print 'What are the mean values read in from the input file?'
    print "Mean Storm Duration is: ", TrialRun.mean_storm, 'hours.'
    print "Interstorm Duration is: ", TrialRun.mean_interstorm, 'hours.'
    print "Storm Depth is: ", TrialRun.mean_storm_depth, 'mm.'
    print "Intensity is: ", TrialRun.mean_intensity, 'mm/hr.'
    print '\n'
    print "Let's see what what the class members are after the first initialization..."
    print "Storm Duration is: ", TrialRun.storm_duration, 'hours.'
    print "Interstorm Duration is: ", TrialRun.interstorm_duration, 'hours.'
    print "Storm Depth is: ", TrialRun.storm_depth, 'mm.'
    print "Intensity is: ", TrialRun.intensity, 'mm/hr.'

    print '\n'
    print 'Now we will update these values using TrialRun.update()'
    TrialRun.update()
    print "Storm Duration is: ", TrialRun.storm_duration, 'hours.'
    print "Interstorm Duration is: ", TrialRun.interstorm_duration, 'hours.'
    print "Storm Depth is: ", TrialRun.storm_depth, 'mm.'
    print "Intensity is: ", TrialRun.intensity, 'mm.'
    
    print '\n'
    print 'Now we are going to generate a time series:'
    TrialRun.get_storm_time_series()
    print TrialRun.storm_time_series


if __name__ == '__main__':
    main()




from .base import ModelGrid
from .hex import HexModelGrid
from .radial import RadialModelGrid
from .raster import RasterModelGrid
from .voronoi import VoronoiDelaunayGrid

from .base import (BAD_INDEX_VALUE, INTERIOR_NODE, FIXED_VALUE_BOUNDARY,
                   FIXED_GRADIENT_BOUNDARY, TRACKS_CELL_BOUNDARY,
                   INACTIVE_BOUNDARY, )
from .create import create_and_initialize_grid



#! /usr/env/python
"""

2D numerical model of diffusion, implemented using ModelGrid.
Provides example of a radial grid.

Last updated GT August 2013

"""

from landlab import RadialModelGrid
import pylab

def main():
    """
    In this simple tutorial example, the main function does all the work: 
    it sets the parameter values, creates and initializes a grid, sets up 
    the state variables, runs the main loop, and cleans up.
    """
    
    import time
    
    start_time = time.time()
    
    # INITIALIZE
    
    # User-defined parameter values
    num_shells=10         # number of radial "shells" in the grid
    #numcols = 30         # not needed for a radial model grid
    dr = 10.0             # grid cell spacing
    kd = 0.01             # diffusivity coefficient, in m2/yr
    uplift_rate = 0.0001  # baselevel/uplift rate, in m/yr
    num_time_steps = 1000 # number of time steps in run
    
    # Derived parameters
    dt = 0.1*dr**2 / kd    # time-step size set by CFL condition
    
    # Create and initialize a radial model grid
    mg = RadialModelGrid(num_shells, dr)
    #mg.initialize(num_shells, dr)
    
    # Set up scalar values
    z = mg.add_zeros('node', 'Elevation')
    #z = mg.create_node_array_zeros('Elevation')            # node elevations
    dzdt = mg.create_node_array_zeros()  # node rate of elevation change
    
    # Get a list of the interior cells
    interior_cells = mg.get_active_cell_node_ids()

    # Display a message
    print( 'Running diffusion_with_radial_model_grid.py' )
    print( 'Time-step size has been set to ' + str( dt ) + ' years.' )

    # RUN
    
    # Main loop
    for i in range(0, num_time_steps):
        
        # Calculate the gradients and sediment fluxes
        g = mg.calculate_gradients_at_active_links(z)
        qs = -kd*g
        
        # Calculate the net deposition/erosion rate at each node
        dqsds = mg.calculate_flux_divergence_at_nodes(qs)
        
        # Calculate the total rate of elevation change
        dzdt = uplift_rate - dqsds
            
        # Update the elevations
        z[interior_cells] = z[interior_cells] + dzdt[interior_cells] * dt

      
    # FINALIZE
    
    # Plot the points, colored by elevation
    import numpy
    maxelev = numpy.amax(z)
    for i in range(mg.number_of_nodes):
        mycolor = str(z[i]/maxelev)
        pylab.plot(mg.node_x[i], mg.node_y[i], 'o', color=mycolor, ms=10)
    
    mg.display_grid()
    
    # Plot the points from the side, with analytical solution
    pylab.figure(3)
    L = num_shells*dr
    xa = numpy.arange(-L, L+dr, dr)
    z_analytical = (uplift_rate/(4*kd))*(L*L-xa*xa)
    pylab.plot(mg.node_x, z, 'o')
    pylab.plot(xa, z_analytical, 'r-')
    pylab.xlabel('Distance from center (m)')
    pylab.ylabel('Height (m)')
    
    pylab.show()
    
    end_time = time.time()
    
    print 'Elapsed time',end_time-start_time
    

if __name__ == "__main__":
    main()




def make_return_array_immutable(func):
    def _wrapped(self, *args, **kwds):
        array = func(self, *args, **kwds)
        immutable_array = array.view()
        immutable_array.flags.writeable = False
        return immutable_array
    return _wrapped



######### soil_moisture_dynamics_driver.py ###########
##
## Soil Moisture Dynamics Model:
## 
##    This Model simulates 'soil moisture dynamics' using following 
## eco-hydrologic components: 
##   
##    1. SM_spatial_DEM_PET.py - Evaluates soil moisture decay every storm
##    2. ET_PriestlyTaylor_Generated.py - Estimates Priestly Taylor potential
##                                                evapotranspiration
##    3. generate_uniform_precip.py - Generate storms
##    4. radiation.py - Estimate influence of aspect on potential -
##                                               evapotranspiration
##    
##    This Model also demonstrates use of cool landlab utilities that import
## a DEM (read_esri_ascii.py) and plot variables on raster grid (imshow.py). 
##
## Sai Nudurupati and E.I. 11/17/2013
#############################################

# Import all required components and tools

from numpy import *
from landlab.components.soil_moisture.SM_spatial_DEM_PET import *
from landlab.components.PET.ET_PriestlyTaylor_Generated import *
from landlab.components.radiation.radiation import *
from landlab.io import read_esri_ascii
import time         # We will need this to check performance of the code

# Point to the input DEM
_DEFAULT_INPUT_FILE_2 = os.path.join(os.path.dirname(__file__),
                                 'saihugoRAW.asc')

""" Create an Soil Moisture Object """
SM = SoilMoisture()

""" Initialize Soil Moisture """
SM.initialize()

""" Set number of storms for the simulation """
SM.set_strms( 5 )
SM._length = SM._iterate_storm

""" Create an instance of Storm Class """
PD = PrecipitationDistribution()

""" Initialize Storm - Create first storm """
PD.initialize()

""" Create an instance of Potential Evapotranspiration (pet) Class """
PET = ET()

""" Initialize potential evaporation class """
PET.initialize()
pet = 0     # local variable initialization

""" Importing Grid and Elevations from DEM """
(RMG,elev) = read_esri_ascii(_DEFAULT_INPUT_FILE_2)

""" Using a dummy variable to capture shape of a multi-dimensional 
    array spread over (all) cells on the raster grid """
dummy = RMG.cell_vector_to_raster(RMG.zeros(centering = 'cell'))

""" Initializing OPTIONAL arrays to store values of variables and plotting """

# following arrays are constant in space but variable in time
""" Store precipitation depth """
P_ = zeros( SM._length, dtype = float )
""" Store effective precipitation depth """ 
Peff_ = zeros( SM._length, dtype = float )
""" Store local time after every storm """
TIME = zeros( SM._length, dtype = float )

# following arrays store data that are variable both in space & time
""" Store soil moisture """
S_ = zeros( (SM._length,dummy.shape[0],dummy.shape[1]), dtype = float )
""" Store actual evapotranspiration """
ETA_ = zeros( (SM._length,dummy.shape[0],dummy.shape[1]), dtype = float )
""" Store drainage """
D_ = zeros( (SM._length,dummy.shape[0],dummy.shape[1]), dtype = float )
""" Store runoff """
Ro_ = zeros( (SM._length,dummy.shape[0],dummy.shape[1]), dtype = float )
""" Store water stress """
WS_ = zeros( (SM._length,dummy.shape[0],dummy.shape[1]), dtype = float )
""" Store radiation factor that records influence of aspect """
RADF_ = zeros( (SM._length,dummy.shape[0],dummy.shape[1]), dtype = float )

""" Initializing 'Initial Soil Moisture' """
SO = 0.5 * ones((dummy.shape))   

""" Create an instance of Radiation Class """
RAD = Radiation()

""" Initialize Radiation class """
RAD.initialize()

# Transfer Elevation data
RAD._Z = RMG.empty()
RAD._Z = elev  

#######
""" Evaluated time taken for the simulation """
Time_start = time.clock()    # Real time for code performance measure
print 'Clock starts ticking now!'

""" Loop through the storms """
for i in range( 0, SM._iterate_storm ):
    if i != 0:
       """ Create a new storm """ 
       PD.update()
    """ Obtain storm data """
    P = PD.storm_depth
    Tb = PD.interstorm_duration
    Tr = PD.storm_duration
    """ Assume fully vegetated field for the time being """
    Fveg = 1.0 #self._soil_Fveg
    
    """ Obtain 'pet' for the storm at half the time between storms.
        Assuming 'pet' will remain constant between the storms"""
    PET.update((TIME[i-1]+(1.0/2.0)*(Tb+Tr)/(24*365.4)))
    pet = max(PET._ETp, 6) # Adjustment of PET to be atleast 6 mm/day
    
    """ Evaluate influence of aspect at the same time when 'pet' is
        obtained.Assuming influence of aspect remains the same 
        between storms """
    RAD.set_current_time( (TIME[i-1]+(1.0/2.0)*(Tb+Tr)/(24*365.4)) )
    RAD.update(RMG)
    RADF_[i] = RMG.cell_vector_to_raster(RAD._Si)

    """ Distribute 'pet' spatially over grid considering aspect influence """
    SM.set_PET((pet*RADF_[i]))    
    

    """ Update Soil Moisture """
    SM.update( P, Tb, Tr, Fveg, SO)
    
    """ Record outputs """
    P_[i] = P
    Peff_[i] = SM.get_Peff()  
    S_[i] = SM.get_S()
    ETA_[i] = SM.get_ET()     
    Ro_[i] = SM.get_Ro()
    D_[i] = SM._D             
    TIME[i] = SM.get_time()      
    WS_[i] = SM.get_WaterStress()
    SO = S_[i]

#######
Time_end = time.clock()                               # Code Performance time
Time_elapsed = Time_end - Time_start                  # Code Performance time
print 'Time Elapsed = ', Time_elapsed, 'secs'        # Code Performance time

""" Plot output - urrently plots soil moisture state at the end of all storms"""

SM.plott( P_, S_[i], ETA_[i], D_[i], Ro_[i], TIME, RMG )


"""
Modules that read/write ModelGrids from various file formats.
"""
from esri_ascii import (read_esri_ascii, read_asc_header)
from esri_ascii import (MissingRequiredKeyError, KeyTypeError,
                        DataSizeError)
